#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from pytext.optimizer.sparsifiers.blockwise_sparsifier import (  # noqa
    BlockwiseMagnitudeSparsifier,
)
